import tap_tester.connections as connections
import tap_tester.runner as runner
from base import TestMixPanelBase


class MixPanelAutomaticFieldsTest(TestMixPanelBase):
    """
    Ensure running the tap with all streams selected and all fields deselected results in the replication of just the
    primary keys and replication keys (automatic fields).
    """

    @staticmethod
    def name():
        return "mix_panel_automatic_fields_test"

    def automatic_fields_test_run(self):
        """
        Verify that for each stream you can get enough data
        when no fields are selected and only the automatic fields are replicated.
        """
        streams_to_test = self.expected_streams()

        conn_id = connections.ensure_connection(self)

        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        test_catalogs_automatic_fields = [catalog for catalog in found_catalogs
                                          if catalog.get('tap_stream_id') in streams_to_test]

        self.perform_and_verify_table_and_field_selection(
            conn_id, test_catalogs_automatic_fields, select_all_fields=False)

        record_count_by_stream = self.run_and_verify_sync(conn_id)
        synced_records = runner.get_records_from_target_output()

        for stream in streams_to_test:
            with self.subTest(stream=stream):

                # expected values
                expected_keys = self.expected_automatic_fields().get(stream)

                # collect actual values
                data = synced_records.get(stream, {})
                record_messages_keys = [set(row['data'].keys())
                                        for row in data.get('messages', [])]

                # Verify that you get some records for each stream
                self.assertGreater(
                    record_count_by_stream.get(stream, 0), 0,
                    msg="The number of records is not over the stream max limit",
                )

                # Verify that only the automatic fields are sent to the target
                for actual_keys in record_messages_keys:
                    self.assertSetEqual(expected_keys, actual_keys)


    def test_standard_auto_fields(self):
        """Automatic fields test for standard server"""
        self.eu_residency = False
        self.automatic_fields_test_run()

    def test_eu_auto_fields(self):
        """Automatic fields test for EU recidency server"""
        self.eu_residency = True
        self.automatic_fields_test_run()



#!/usr/bin/env python3

import sys
import json
import argparse
from datetime import datetime, timedelta, date
import singer
from singer import metadata, utils
from singer.utils import strptime_to_utc, strftime
from tap_mixpanel.client import MixpanelClient
from tap_mixpanel.discover import discover
from tap_mixpanel.sync import sync

LOGGER = singer.get_logger()

REQUEST_TIMEOUT = 300
REQUIRED_CONFIG_KEYS = [
    'project_timezone',
    'api_secret',
    'attribution_window',
    'start_date',
    'user_agent'
]


def do_discover(client, properties_flag):
    LOGGER.info('Starting discover')
    catalog = discover(client, properties_flag)
    json.dump(catalog.to_dict(), sys.stdout, indent=2)
    LOGGER.info('Finished discover')


@singer.utils.handle_top_exception(LOGGER)
def main():
    parsed_args = singer.utils.parse_args(REQUIRED_CONFIG_KEYS)

    start_date = parsed_args.config['start_date']
    # Set request timeout to config param `request_timeout` value.
    # If value is 0, "0", "" or not passed then it sets default to 300 seconds.
    config_request_timeout = parsed_args.config.get('request_timeout')
    if config_request_timeout and float(config_request_timeout):
        request_timeout = float(config_request_timeout)
    else:
        request_timeout = REQUEST_TIMEOUT

    start_dttm = strptime_to_utc(start_date)
    now_dttm = utils.now()
    if parsed_args.config.get('end_date'):
        now_dttm = strptime_to_utc(parsed_args.config.get('end_date'))
    delta_days = (now_dttm - start_dttm).days
    if delta_days >= 365:
        delta_days = 365
        start_date = strftime(now_dttm - timedelta(days=delta_days))
        LOGGER.warning("start_date greater than 1 year maxiumum for API.")
        LOGGER.warning("Setting start_date to 1 year ago, %s", start_date)

    #Check support for EU endpoints
    if str(parsed_args.config.get('eu_residency')).lower() == "true":
        api_domain = "eu.mixpanel.com"
    else:
        api_domain = "mixpanel.com"

    with MixpanelClient(parsed_args.config['api_secret'],
                        api_domain,
                        request_timeout,
                        parsed_args.config['user_agent']) as client:

        state = {}
        if parsed_args.state:
            state = parsed_args.state

        config = parsed_args.config
        properties_flag = config.get('select_properties_by_default')

        if parsed_args.discover:
            client.__api_domain = api_domain
            do_discover(client, properties_flag)
        elif parsed_args.catalog:
            sync(client=client,
                 config=config,
                 catalog=parsed_args.catalog,
                 state=state,
                 start_date=start_date)


if __name__ == '__main__':
    main()



import unittest
from unittest import mock
import requests
from tap_mixpanel import client

# mock responce
REQUEST_TIMEOUT = 300


class Mockresponse:
    """
    Mocked standard HTTPResponse to test error handling.
    """
    def __init__(self, resp, status_code, content=[""], headers=None, raise_error=False, text={}):
        self.json_data = resp
        self.status_code = status_code
        self.content = content
        self.headers = headers
        self.raise_error = raise_error
        self.text = text
        self.reason = "error"

    def prepare(self):
        return (self.json_data, self.status_code, self.content, self.headers, self.raise_error)

    def raise_for_status(self):
        if not self.raise_error:
            return self.status_code

        raise requests.HTTPError("mock sample message")

    def json(self):
        return self.text

# Mock response for timezone related error messages
def get_mock_http_response(content, status_code):
    response = requests.Response()
    response.status_code = status_code
    response.headers = {}
    response._content = content.encode()
    return response

@mock.patch('time.sleep', return_value=None) # Mock time.sleep to reduce the time
class TestMixpanelErrorHandling(unittest.TestCase):

    def mock_send_400(*args, **kwargs):
        return Mockresponse("", 400, raise_error=True)

    def mock_400_different_timezone(*args, **kwargs):
        content = " to_date cannot be later than today"
        return get_mock_http_response(content, 400)

    def mock_send_401(*args, **kwargs):
        return Mockresponse("", 401, raise_error=True)

    def mock_send_402(*args, **kwargs):
        return Mockresponse("", 402, raise_error=True)

    def mock_send_403(*args, **kwargs):
        return Mockresponse("", 403, raise_error=True)

    def mock_send_404(*args, **kwargs):
        return Mockresponse("", 404, raise_error=True)

    def mock_send_429(*args, **kwargs):
        return Mockresponse("", 429, raise_error=True)

    def mock_send_500(*args, **kwargs):
        return Mockresponse("", 500, raise_error=True)

    def mock_send_501(*args, **kwargs):
        return Mockresponse("", 501, raise_error=True)

    def mock_send_error(*args, **kwargs):
        content = '{"error": "Resource not found error message from API response field \'error\'."}'
        return get_mock_http_response(content, 404)

    def mock_send_message(*args, **kwargs):
        content = '{"message": "Resource not found error message from API response field \'message\'."}'
        return get_mock_http_response(content, 404)

    @mock.patch("requests.Session.request", side_effect=mock_send_400)
    def test_request_with_handling_for_400_exception_handling(self, mock_send_400, mock_sleep):
        """
        Test that `perform_request` method handle 400 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelBadRequestError as e:
            expected_error_message = "HTTP-error-code: 400, Error: A validation exception has occurred.(Please verify your credentials.)"
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_400_different_timezone)
    def test_request_with_handling_for_400_for_different_timezone(self, mock_400_different_timezone, mock_sleep):
        """
        Test that `perform_request` method handle 400 error with proper message for different timezone
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelBadRequestError as e:
            expected_error_message = "HTTP-error-code: 400, Error: A validation exception has occurred. Please validate the timezone with the MixPanel UI under project settings."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request")
    def test_request_with_handling_for_400_timeout_error_handling(self, mock_request, mock_sleep):
        """
        Test that `perform_request` method handle 400 error with timeout error message in case of `error` field in response
        """
        error = {"request": "/api/2.0/engage/revenue?from_date=2020-02-01&to_date=2020-03-01", "error": "Timeout Error."}
        mock_request.return_value = Mockresponse("", 400, raise_error=True, text=error)
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelBadRequestError as e:
            expected_error_message = "HTTP-error-code: 400, Error: Timeout Error.(Please verify your credentials.)"
            # Verifying the message formed for the timeout error
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_401)
    def test_request_with_handling_for_401_exception_handling(self, mock_send_401, mock_sleep):
        """
        Test that `perform_request` method handle 401 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelUnauthorizedError as e:
            expected_error_message = "HTTP-error-code: 401, Error: Invalid authorization credentials."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_402)
    def test_request_with_handling_for_402_exception_handling(self, mock_send_402, mock_sleep):
        """
        Test that `perform_request` method handle 402 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelPaymentRequiredError as e:
            expected_error_message = "HTTP-error-code: 402, Error: Your current plan does not allow API calls. Payment is required to complete the operation."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_403)
    def test_request_with_handling_for_403_exception_handling(self, mock_send_403, mock_sleep):
        """
        Test that `perform_request` method handle 403 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelForbiddenError as e:
            expected_error_message = "HTTP-error-code: 403, Error: User does not have permission to access the resource."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_404)
    def test_request_with_handling_for_404_exception_handling(self, mock_send_404, mock_sleep):
        """
        Test that `perform_request` method handle 404 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelNotFoundError as e:
            expected_error_message = "HTTP-error-code: 404, Error: The resource you have specified cannot be found."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_429)
    def test_request_with_handling_for_429_exception_handling(self, mock_send_429, mock_sleep):
        """
        Test that `perform_request` method handle 429 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.Server429Error as e:
            expected_error_message = "HTTP-error-code: 429, Error: The API rate limit for your organisation/application pairing has been exceeded."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_500)
    def test_request_with_handling_for_500_exception_handling(self, mock_send_500, mock_sleep):
        """
        Test that `perform_request` method handle 500 error with proper message
        """
        with self.assertRaises(client.MixpanelInternalServiceError):
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')

    @mock.patch("requests.Session.request", side_effect=mock_send_501)
    def test_request_with_handling_for_501_exception_handling(self, mock_send_501, mock_sleep):
        """
        Test that `perform_request` method handle 501 error with proper message
        """
        with self.assertRaises(client.Server5xxError):
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')

    @mock.patch("requests.Session.request", side_effect=mock_send_error)
    def test_request_with_handling_for_404_exception_handling_error(self, mock_send_error, mock_sleep):
        '''
            Verify that if 'error' field is present in API response then it should be used as error message.
        '''
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelNotFoundError as e:
            expected_error_message = "HTTP-error-code: 404, Error: Resource not found error message from API response field 'error'."
            # Verifying the message retrived from 'error' field of API response
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_message)
    def test_request_with_handling_for_404_exception_handling_message(self, mock_send_message, mock_sleep):
        '''
            Verify that if 'message' field is present in API response then it should be used as error message.
        '''
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.perform_request('GET')
        except client.MixpanelNotFoundError as e:
            expected_error_message = "HTTP-error-code: 404, Error: Resource not found error message from API response field 'message'."
            # Verifying the message retrived from 'message' field of API response
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.get", side_effect=mock_send_400)
    def test_check_access_with_handling_for_400_exception_handling(self, mock_send_400, mock_sleep):
        """
        Test that `check_access` method handle 404 error with proper message
        """
        try:
            tap_stream_id = "tap_mixpanel"
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelBadRequestError as e:
            expected_error_message = "HTTP-error-code: 400, Error: A validation exception has occurred.(Please verify your credentials.)"
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.get", side_effect=mock_400_different_timezone)
    def test_check_access_with_handling_for_400_for_different_timezone(self, mock_400_different_timezone, mock_sleep):
        """
        Test that `check_access` method handle 404 error with proper message for different timezone
        """
        try:
            tap_stream_id = "tap_mixpanel"
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelBadRequestError as e:
            expected_error_message = "HTTP-error-code: 400, Error: A validation exception has occurred. Please validate the timezone with the MixPanel UI under project settings."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.get")
    def test_check_access_with_handling_for_400_timeout_error_handling(self, mock_request, mock_sleep):
        """
        Test that `check_access` method handle 404 error with timeout error message in case of `error` field in response
        """
        error = {"request": "/api/2.0/engage/revenue?from_date=2020-02-01&to_date=2020-03-01", "error": "Timeout Error."}
        mock_request.return_value = Mockresponse("", 400, raise_error=True, text=error)
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelBadRequestError as e:
            expected_error_message = "HTTP-error-code: 400, Error: Timeout Error.(Please verify your credentials.)"
            # Verifying the message formed for the timeout error
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_401)
    def test_check_access_with_handling_for_401_exception_handling(self, mock_send_401, mock_sleep):
        """
        Test that `check_access` method handle 401 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelUnauthorizedError as e:
            expected_error_message = "HTTP-error-code: 401, Error: Invalid authorization credentials."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_403)
    def test_check_access_with_handling_for_403_exception_handling(self, mock_send_403, mock_sleep):
        """
        Test that `check_access` method handle 403 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelForbiddenError as e:
            expected_error_message = "HTTP-error-code: 403, Error: User does not have permission to access the resource."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_404)
    def test_check_access_with_handling_for_404_exception_handling(self, mock_send_404, mock_sleep):
        """
        Test that `check_access` method handle 404 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelNotFoundError as e:
            expected_error_message = "HTTP-error-code: 404, Error: The resource you have specified cannot be found."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_429)
    def test_check_access_with_handling_for_429_exception_handling(self, mock_send_429, mock_sleep):
        """
        Test that `check_access` method handle 429 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.Server429Error as e:
            expected_error_message = "HTTP-error-code: 429, Error: The API rate limit for your organisation/application pairing has been exceeded."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_500)
    def test_check_access_with_handling_for_500_exception_handling(self, mock_send_500, mock_sleep):
        """
        Test that `check_access` method handle 500 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelInternalServiceError as e:
            expected_error_message = "HTTP-error-code: 500, Error: Server encountered an unexpected condition that prevented it from fulfilling the request."
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_501)
    def test_check_access_with_handling_for_501_exception_handling(self, mock_send_501, mock_sleep):
        """
        Test that `check_access` method handle 501 error with proper message
        """
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelError as e:
            expected_error_message = "HTTP-error-code: 501, Error: Unknown Error"
            # Verifying the message formed for the custom exception
            self.assertEqual(str(e), expected_error_message)


    @mock.patch("requests.Session.request", side_effect=mock_send_error)
    def test_check_access_with_handling_for_404_exception_handling_error(self, mock_send_error, mock_sleep):
        '''
            Verify that if 'error' field is present in API response then it should be used as error message.
        '''
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelNotFoundError as e:
            expected_error_message = "HTTP-error-code: 404, Error: Resource not found error message from API response field 'error'."
            # Verifying the message retrived from 'error' field of API response
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=mock_send_message)
    def test_check_access_with_handling_for_404_exception_handling_message(self, mock_send_message, mock_sleep):
        '''
            Verify that if 'message' field is present in API response then it should be used as error message.
        '''
        try:
            mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
            mock_client.check_access()
        except client.MixpanelNotFoundError as e:
            expected_error_message = "HTTP-error-code: 404, Error: Resource not found error message from API response field 'message'."
            # Verifying the message retrived from 'message' field of API response
            self.assertEqual(str(e), expected_error_message)

    @mock.patch("requests.Session.request", side_effect=requests.exceptions.Timeout)
    def test_check_access_handle_timeout_error(self, mock_request, mock_sleep):
        '''
        Check whether the request backoffs properly for `check_access` method for 5 times in case of Timeout error.
        '''
        mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
        with self.assertRaises(client.ReadTimeoutError):
            mock_client.check_access()

        # Verify that requests.Session.request is called 5 times
        self.assertEqual(mock_request.call_count, 5)




class TestMixpanelConnectionResetErrorHandling(unittest.TestCase):

    @mock.patch("requests.Session.request", side_effect=requests.models.ProtocolError)
    def test_check_access_handle_timeout_error(self, mock_request):
        '''
        Check whether the request backoffs properly for `check_access` method for 5 times in case of Timeout error.
        '''
        mock_client = client.MixpanelClient(api_secret="mock_api_secret", api_domain="mock_api_domain", request_timeout=REQUEST_TIMEOUT)
        with self.assertRaises(requests.models.ProtocolError):
            mock_client.check_access()

        # Verify that requests.Session.request is called 5 times
        self.assertEqual(mock_request.call_count, 5)



from collections.abc import Generator
from unittest import mock
from unittest.mock import patch

import backoff
import requests
import requests_mock
from pytest import raises
from tap_mixpanel import client
from tap_mixpanel.client import (ReadTimeoutError, Server5xxError,
                                 Server429Error, MixpanelInternalServiceError)
from tests.configuration.fixtures import mixpanel_client


@mock.patch('time.sleep', return_value=None)
def test_request_export_backoff_on_timeout(mock_sleep, mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'http://test.com',
                  exc=requests.exceptions.Timeout('Timeout on request'))

        with raises(ReadTimeoutError) as ex:
            for record in mixpanel_client.request_export('GET', url='http://test.com'):
                pass
        # Assert backoff retry count as expected
        assert mock_sleep.call_count == client.BACKOFF_MAX_TRIES_REQUEST - 1


@mock.patch('time.sleep', return_value=None)
def test_request_export_backoff_on_remote_timeout(mock_sleep, mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'http://test.com', text=None, status_code=504)
        result = mixpanel_client.request_export('GET', url='http://test.com')

        with raises(Server5xxError) as ex:
            for record in result:
                pass
        # Assert backoff retry count as expected
        assert mock_sleep.call_count == client.BACKOFF_MAX_TRIES_REQUEST - 1


@mock.patch('time.sleep', return_value=None)
def test_perform_request_backoff_on_remote_timeout_429(mock_sleep, mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'http://test.com', text=None,
                  content=b'error', status_code=429)

        with raises(Server429Error) as ex:
            result = mixpanel_client.perform_request(
                'GET', url='http://test.com')
            for record in result:
                pass
        # Assert backoff retry count as expected
        assert mock_sleep.call_count == client.BACKOFF_MAX_TRIES_REQUEST - 1


@mock.patch('time.sleep', return_value=None)
def test_perform_request_backoff_on_remote_timeout_500(mock_sleep, mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'http://test.com', text=None, status_code=500)

        with raises(MixpanelInternalServiceError) as ex:
            result = mixpanel_client.perform_request(
                'GET', url='http://test.com')

            for record in result:
                pass
        # Assert backoff retry count as expected
        assert mock_sleep.call_count == client.BACKOFF_MAX_TRIES_REQUEST - 1


@mock.patch('time.sleep', return_value=None)
def test_check_access_backoff_on_remote_timeout_429(mock_sleep, mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'https://mixpanel.com/api/2.0/engage',
                  content=b'error', text=None, status_code=429)

        with raises(Server429Error) as ex:
            result = mixpanel_client.check_access()
        # Assert backoff retry count as expected
        assert mock_sleep.call_count == 5 - 1


@mock.patch('time.sleep', return_value=None)
def test_check_access_backoff_on_remote_timeout_500(mock_sleep, mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'https://mixpanel.com/api/2.0/engage',
                  content=b'error', text=None, status_code=500)

        with raises(MixpanelInternalServiceError) as ex:
            result = mixpanel_client.check_access()
        # Assert backoff retry count as expected
        assert mock_sleep.call_count == 5 - 1


@mock.patch('time.sleep', return_value=None)
def test_request_backoff_on_timeout(mock_sleep, mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'http://test.com',
                  exc=requests.exceptions.Timeout('Timeout on request'))

        with raises(ReadTimeoutError) as ex:
            result = mixpanel_client.request('GET', url='http://test.com')
        # Assert backoff retry count as expected
        assert mock_sleep.call_count == client.BACKOFF_MAX_TRIES_REQUEST - 1


def test_request_returns_json(mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'http://test.com', json={'a': 'b'})
        result = mixpanel_client.request('GET', url='http://test.com')
        assert result == {'a': 'b'}


def test_request_export_returns_generator(mixpanel_client):
    with requests_mock.Mocker() as m:
        m.request('GET', 'http://test.com', json={'a': 'b'})
        result = mixpanel_client.request_export('GET', url='http://test.com')
        assert isinstance(result, Generator)



import base64
import io

import backoff
import jsonlines
import requests
import singer
from requests.exceptions import ConnectionError, Timeout
from requests.models import ProtocolError
from singer import metrics

LOGGER = singer.get_logger()

BACKOFF_MAX_TRIES_REQUEST = 7
REQUEST_TIMEOUT = 300


class ReadTimeoutError(Exception):
    pass


class Server5xxError(Exception):
    pass


class Server429Error(Exception):
    pass


class MixpanelError(Exception):
    pass


class MixpanelBadRequestError(MixpanelError):
    pass


class MixpanelUnauthorizedError(MixpanelError):
    pass


class MixpanelPaymentRequiredError(MixpanelError):
    pass


class MixpanelNotFoundError(MixpanelError):
    pass


class MixpanelForbiddenError(MixpanelError):
    pass


class MixpanelInternalServiceError(Server5xxError):
    pass


ERROR_CODE_EXCEPTION_MAPPING = {
    400: {
        "raise_exception": MixpanelBadRequestError,
        "message": "A validation exception has occurred."
    },
    401: {
        "raise_exception": MixpanelUnauthorizedError,
        "message": "Invalid authorization credentials."
    },
    402: {
        "raise_exception": MixpanelPaymentRequiredError,
        "message": "Your current plan does not allow API calls. Payment is required to complete the operation."
    },
    403: {
        "raise_exception": MixpanelForbiddenError,
        "message": "User does not have permission to access the resource."
    },
    404: {
        "raise_exception": MixpanelNotFoundError,
        "message": "The resource you have specified cannot be found."
    },
    429: {
        "raise_exception": Server429Error,
        "message": "The API rate limit for your organisation/application pairing has been exceeded."
    },
    500: {
        "raise_exception": MixpanelInternalServiceError,
        "message": "Server encountered an unexpected condition that prevented it from fulfilling the request."
    }
}

def raise_for_error(response):
    LOGGER.error('ERROR %s: %s, REASON: %s', response.status_code,
                                             response.text, 
                                             response.reason)
    try:
        response_json = response.json()
    except Exception:
        response_json = {}
    error_code = response.status_code
    error_message = response_json.get(
        "error", response_json.get(
            "message", ERROR_CODE_EXCEPTION_MAPPING.get(
                error_code, {}).get(
                    "message", "Unknown Error")))

    # if response text contains something unusual error of to_date then provide helper message of timezone mismatch
    # E.g error: to_date cannot be later than today
    if error_code == 400:
        if "to_date" in response.text:
            error_message += " Please validate the timezone with the MixPanel UI under project settings."
        else:
            error_message = '{}(Please verify your credentials.)'.format(error_message)

    message = "HTTP-error-code: {}, Error: {}".format(error_code, error_message)

    exc = ERROR_CODE_EXCEPTION_MAPPING.get(
        error_code, {}).get("raise_exception", MixpanelError)
    raise exc(message) from None


class MixpanelClient(object):
    def __init__(self,
                 api_secret,
                 api_domain,
                 request_timeout,
                 user_agent=None):
        self.__api_secret = api_secret
        self.__api_domain = api_domain
        self.__request_timeout = request_timeout
        self.__user_agent = user_agent
        self.__session = requests.Session()
        self.__verified = False
        self.disable_engage_endpoint = False

    def __enter__(self):
        self.__verified = self.check_access()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.__session.close()

    @backoff.on_exception(backoff.expo,
                          (Server5xxError, Server429Error, ReadTimeoutError, ConnectionError, Timeout, ProtocolError),
                          max_tries=5,
                          factor=2)
    def check_access(self):
        if self.__api_secret is None:
            raise Exception('Error: Missing api_secret in tap config.json.')
        headers = {}
        # Endpoint: simple API call to return a single record (org settings) to test access
        url = 'https://{}/api/2.0/engage'.format(self.__api_domain)
        if self.__user_agent:
            headers['User-Agent'] = self.__user_agent
        headers['Accept'] = 'application/json'
        headers['Authorization'] = 'Basic {}'.format(
            str(base64.urlsafe_b64encode(self.__api_secret.encode("utf-8")), "utf-8"))

        try:
            response = self.__session.get(
                url=url,
                timeout=self.__request_timeout, # Request timeout parameter
                headers=headers)
        except requests.exceptions.Timeout as err:
            LOGGER.error('TIMEOUT ERROR: %s',str(err))
            raise ReadTimeoutError

        if response.status_code == 402:
            # 402 Payment Requirement does not indicate a permissions or authentication error
            self.disable_engage_endpoint = True
            LOGGER.warning('Mixpanel returned a 402 from the Engage API. Engage stream will be skipped.')
            return True
        elif response.status_code != 200:
            LOGGER.error('Error status_code = {}'.format(response.status_code))
            raise_for_error(response)
        else:
            return True

    @backoff.on_exception(
        backoff.expo,
        (Server5xxError, Server429Error, ReadTimeoutError, ConnectionError, Timeout, ProtocolError),
        max_tries=BACKOFF_MAX_TRIES_REQUEST,
        factor=3,
        logger=LOGGER)
    def perform_request(self,
                        method,
                        url=None,
                        params=None,
                        json=None,
                        stream=False,
                        **kwargs):
        try:
            response = self.__session.request(method=method,
                                              url=url,
                                              params=params,
                                              json=json,
                                              stream=stream,
                                              timeout=self.__request_timeout, # Request timeout parameter
                                              **kwargs)

            if response.status_code > 500:
                raise Server5xxError()

            if response.status_code != 200:
                raise_for_error(response)
            return response
        except requests.exceptions.Timeout as err:
            LOGGER.error('TIMEOUT ERROR: %s',str(err))
            raise ReadTimeoutError(err)

    def request(self, method, url=None, path=None, params=None, json=None, **kwargs):
        if not self.__verified:
            self.__verified = self.check_access()

        if url and path:
            url = '{}/{}'.format(url, path)
        elif path and not url:
            url = 'https://{}/api/2.0/{}'.format(self.__api_domain, path)

        if 'endpoint' in kwargs:
            endpoint = kwargs['endpoint']
            del kwargs['endpoint']
        else:
            endpoint = None

        if 'headers' not in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['Accept'] = 'application/json'

        if self.__user_agent:
            kwargs['headers']['User-Agent'] = self.__user_agent

        if method == 'POST':
            kwargs['headers']['Content-Type'] = 'application/json'

        kwargs['headers']['Authorization'] = 'Basic {}'.format(
            str(base64.urlsafe_b64encode(self.__api_secret.encode("utf-8")), "utf-8"))
        with metrics.http_request_timer(endpoint) as timer:
            response = self.perform_request(method=method,
                                            url=url,
                                            params=params,
                                            json=json,
                                            **kwargs)

            timer.tags[metrics.Tag.http_status_code] = response.status_code

        response_json = response.json()
        return response_json

    def request_export(self, method, url=None, path=None, params=None, json=None, **kwargs):
        if not self.__verified:
            self.__verified = self.check_access()

        if url and path:
            url = '{}/{}'.format(url, path)
        elif path and not url:
            url = 'https://{}/api/2.0/{}'.format(self.__api_domain, path)

        if 'endpoint' in kwargs:
            endpoint = kwargs['endpoint']
            del kwargs['endpoint']
        else:
            endpoint = 'export'

        if 'headers' not in kwargs:
            kwargs['headers'] = {}

        kwargs['headers']['Accept'] = 'application/json'

        if self.__user_agent:
            kwargs['headers']['User-Agent'] = self.__user_agent

        if method == 'POST':
            kwargs['headers']['Content-Type'] = 'application/json'

        kwargs['headers']['Authorization'] = 'Basic {}'.format(
            str(base64.urlsafe_b64encode(self.__api_secret.encode("utf-8")), "utf-8"))
        with metrics.http_request_timer(endpoint) as timer:
            response = self.perform_request(method=method,
                                            url=url,
                                            params=params,
                                            json=json,
                                            stream=True,
                                            **kwargs)
            timer.tags[metrics.Tag.http_status_code] = response.status_code

            # export endpoint returns jsonl results;
            #  other endpoints return json with array of results
            #  jsonlines reference: https://jsonlines.readthedocs.io/en/latest/
            reader = jsonlines.Reader(response.iter_lines())
            for record in reader.iter(allow_none=True, skip_empty=True):
                yield record



from tap_mixpanel.transform import transform_event_times
import pytz
import unittest
from datetime import datetime

UTC = pytz.utc


class TestTransformEventTimes(unittest.TestCase):

    def test_utc_now(self):
        
        input_time = datetime.utcnow()
        
        record = {"time": input_time.timestamp()}
        project_timezone = "UTC"
        
        actual = transform_event_times(record, project_timezone)
        expected = {"time": input_time.astimezone(UTC).strftime("%04Y-%m-%dT%H:%M:%S.000000Z")}
        
        self.assertEqual(expected, actual)
        
        
    def test_eastern_time(self):
        project_timezone = "US/Eastern"
        EASTERN = pytz.timezone(project_timezone)
        # This gives us 2021-08-12T11:00:00-4:00
        input_time = EASTERN.localize(datetime(2021, 8, 12, 11, 0, 0))
        
        record = {"time": input_time.timestamp()}
        actual = transform_event_times(record, project_timezone)
        
        expected = {"time": input_time.astimezone(UTC).strftime("%04Y-%m-%dT%H:%M:%S.000000Z")}

        self.assertEqual(expected, actual)


import datetime
import pytz
import singer
from singer.utils import strftime
from singer import Transformer

LOGGER = singer.get_logger()

# De-nest properties for engage and export endpoints
def denest_properties(record, properties_node):
    new_record = record
    properties = record.get(properties_node)
    if properties:
        for key, val in record[properties_node].items():
            if key[0:1] == '$':
                new_key = 'mp_reserved_{}'.format(key[1:])
                # change this to regex
            else:
                new_key = key
            new_record[new_key] = val
        new_record.pop(properties_node, None)
    return new_record


# Time conversion from $time integer using project_timezone
# Reference: https://help.mixpanel.com/hc/en-us/articles/115004547203-Manage-Timezones-for-Projects-in-Mixpanel#exporting-data-from-mixpanel
def transform_event_times(record, project_timezone):
    new_record = record
    timezone = pytz.timezone(project_timezone)

    # Create beginning_datetime: beginning of epoch time in project timezone
    naive_time = datetime.time(0, 0)
    date = datetime.date(1970, 1, 1)
    naive_datetime = datetime.datetime.combine(date, naive_time)
    # Move 1970-01-01T00:00:00Z to Mixpanel project timezone.
    # For example, if Mixpanel timezone is in Eastern Time (UTC-5:00) then it calculates 1969-12-31T19:00:00-5:00.
    beginning_datetime = pytz.utc.localize(naive_datetime).astimezone(timezone)

    # Get integer time
    time_int = int(record.get('time'))

    # Create new_time_utc by adding seconds to beginning_datetime, normalizing,
    #   and converting to string
    add_seconds = datetime.timedelta(seconds=time_int)
    new_time = beginning_datetime + add_seconds

    # 'normalize' accounts for daylight savings time
    new_time_utc_str = strftime(timezone.normalize(new_time).astimezone(pytz.utc))
    new_record['time'] = new_time_utc_str

    return new_record

def transform_datetime(this_dttm):
        with Transformer() as transformer:
            new_dttm = transformer._transform_datetime(this_dttm)
        return new_dttm


# Remove leading $ from engage $distinct_id
def transform_engage(record):
    new_record = record
    distinct_id = record.get('$distinct_id')
    new_record['distinct_id'] = distinct_id
    new_record.pop('$distinct_id', None)
    return new_record


# Funnels: combine parent record with each date record
def transform_funnels(record, parent_record):
    record.update(parent_record)
    return record


# Cohort Members: provide all distinct_id's for each cohort_id
def transform_cohort_members(record, parent_record):
    cohort_id = parent_record.get('id')
    distinct_id = record.get('$distinct_id')
    new_record = {}
    new_record['distinct_id'] = distinct_id
    new_record['cohort_id'] = cohort_id
    return new_record


# Run other transforms, as needed: denest_list_nodes, transform_conversation_parts
def transform_record(record, stream_name, project_timezone, parent_record=None):
    if stream_name == 'engage':
        trans_json = transform_engage(record)
        new_record = denest_properties(trans_json, '$properties')
    elif stream_name == 'export':
        denested_json = denest_properties(record, 'properties')
        new_record = transform_event_times(denested_json, project_timezone)
    elif stream_name == 'funnels':
        new_record = transform_funnels(record, parent_record)
    elif stream_name == 'cohort_members':
        new_record = transform_cohort_members(record, parent_record)
    else:
        new_record = record

    return new_record



"""
This module defines the stream classes and their individual sync logic.
"""

import json
import math
from datetime import datetime, timedelta

import pytz
import singer
from singer import Transformer, metadata, metrics, utils
from singer.utils import strptime_to_utc

from tap_mixpanel.client import MixpanelClient
from tap_mixpanel.transform import transform_record, transform_datetime

LOGGER = singer.get_logger()


class MixPanel:
    """
    A base class representing singer streams.
    :param client: The API client used to extract records from external source
    """
    tap_stream_id = None
    replication_method = None
    replication_keys = []
    key_properties = []
    to_replicate = True
    date_dictionary = False
    path = None
    params = {}
    parent = None
    data_key = None
    bookmark_query_field_from = None
    bookmark_query_field_to = None
    pagination = False
    parent_path = None
    parent_id_field = None
    url = "https://mixpanel.com/api/2.0"

    def __init__(self, client: MixpanelClient):
        self.client = client

    def write_schema(self, catalog, stream_name):
        stream = catalog.get_stream(stream_name)
        schema = stream.schema.to_dict()
        try:
            singer.write_schema(stream_name, schema, stream.key_properties)
        except OSError as err:
            LOGGER.error("OS Error writing schema for: %s",stream_name)
            raise err

    def get_bookmark(self, state, stream, default):
        if (state is None) or ("bookmarks" not in state):
            return default
        return state.get("bookmarks", {}).get(stream, default)

    def write_bookmark(self, state, stream, value):
        if "bookmarks" not in state:
            state["bookmarks"] = {}
        state["bookmarks"][stream] = value
        LOGGER.info("Write state for stream: %s, value: %s", stream, value)
        singer.write_state(state)

    def process_records(
        self,  # pylint: disable=too-many-branches
        catalog,
        stream_name,
        records,
        time_extracted,
        bookmark_field=None,
        max_bookmark_value=None,
        last_datetime=None,
    ):
        stream = catalog.get_stream(stream_name)
        schema = stream.schema.to_dict()
        stream_metadata = metadata.to_map(stream.metadata)

        with metrics.record_counter(stream_name) as counter:
            for record in records:
                # Transform record for Singer.io
                with Transformer() as transformer:
                    try:
                        transformed_record = transformer.transform(
                            record, schema, stream_metadata
                        )
                    except Exception as err:
                        LOGGER.error("Error: %s",str(err))
                        LOGGER.error("For schema: %s",
                            json.dumps(schema, sort_keys=True, indent=2)
                        )
                        raise err

                    # Reset max_bookmark_value to new value if higher
                    if transformed_record.get(bookmark_field):
                        if max_bookmark_value is None or transformed_record[
                            bookmark_field
                        ] > transform_datetime(max_bookmark_value):
                            max_bookmark_value = transformed_record[bookmark_field]

                    if bookmark_field and (bookmark_field in transformed_record):
                        last_dttm = transform_datetime(last_datetime)
                        bookmark_dttm = transform_datetime(
                            transformed_record[bookmark_field]
                        )
                        # Keep only records whose bookmark is after the last_datetime
                        if bookmark_dttm >= last_dttm:
                            singer.write_record(
                                stream_name,
                                transformed_record,
                                time_extracted=time_extracted,
                            )
                            counter.increment()
                    else:
                        singer.write_record(
                            stream_name,
                            transformed_record,
                            time_extracted=time_extracted,
                        )
                        counter.increment()

            return max_bookmark_value, counter.value

    def get_and_transform_records(
            self, querystring, project_timezone, max_bookmark_value, catalog, last_datetime, endpoint_total,
            limit, total_records, parent_total, record_count, page, offset, parent_record, date_total):
        """
        Get the records using the client get request and transform it using transform_records
        and return the max_bookmark_value
        """

        session_id = None
        data = self.client.request(
            method='GET',
            url=self.url,
            path=self.path,
            params=querystring,
            endpoint=self.tap_stream_id)

        # time_extracted: datetime when the data was extracted from the API
        time_extracted = utils.now()
        full_url = '{}/{}{}'.format(
            self.url,
            self.path,
            '?{}'.format(querystring) if querystring else '')
        if not data:
            LOGGER.info('No data for URL: %s',full_url)
            # No data results
        else:  # has data
            # Transform data with transform_json from transform.py
            # The data_key identifies the array/list of records below the <root> element
            # LOGGER.info('data = {}'.format(data)) # TESTING, comment out
            transformed_data = []  # initialize the record list

            # Endpoints: funnels, revenue return results as dictionary for each date
            # Standardize results to a list/array
            if self.date_dictionary and self.data_key in data:
                results = {}
                results_list = []
                for key, val in data[self.data_key].items():
                    # skip $overall summary
                    if key != '$overall':
                        val['date'] = key
                        val['datetime'] = '{}T00:00:00Z'.format(key)
                        results_list.append(val)
                results[self.data_key] = results_list
                data = results

            # Cohorts endpoint returns results as a list/array (no data_key)
            # All other endpoints have a data_key
            if not self.data_key:
                self.data_key = 'results'
                new_data = {
                    'results': data
                }
                data = new_data

            transformed_data = []
            # Loop through result records
            for record in data[self.data_key]:
                # transform record and append to transformed_data array
                transformed_record = transform_record(record,
                                                      self.tap_stream_id,
                                                      project_timezone,
                                                      parent_record)
                transformed_data.append(transformed_record)

                # Check for missing keys
                for key in self.key_properties:
                    val = transformed_record.get(key)
                    if not val:
                        LOGGER.error('Error: Missing Key')
                        raise 'Missing Key'

                # End data record loop

            if not transformed_data:
                 LOGGER.info('No transformed data for data = %s', data)
                # No transformed data results
            else:  # has transformed data
                # Process records and get the max_bookmark_value and record_count
                max_bookmark_value, record_count = self.process_records(
                    catalog=catalog,
                    stream_name=self.tap_stream_id,
                    records=transformed_data,
                    time_extracted=time_extracted,
                    bookmark_field=next(iter(self.replication_keys), None),
                    max_bookmark_value=max_bookmark_value,
                    last_datetime=last_datetime)
                LOGGER.info('Stream %s, batch processed %s records', self.tap_stream_id, record_count)

                # set total_records and pagination fields
                if page == 0:
                    if isinstance(data, dict):
                        total_records = data.get('total', record_count)
                    else:
                        total_records = record_count
                parent_total = parent_total + record_count
                date_total = date_total + record_count
                endpoint_total = endpoint_total + record_count
                if isinstance(data, dict):
                    session_id = data.get('session_id', None)

                # to_rec: to record; ending record for the batch page
                if self.pagination:
                    to_rec = offset + limit
                    if to_rec > total_records:
                        to_rec = total_records
                else:
                    to_rec = record_count

                LOGGER.info('Synced Stream: %s, page: %s, %s to %s of total: %s',
                            self.tap_stream_id,
                            page,
                            offset,
                            to_rec,
                            total_records)
                # End has transformed data
            # End has data results

        # Pagination: increment the offset by the limit (batch-size) and page
        offset = offset + limit
        page = page + 1
        return parent_total, date_total, offset, page, session_id, endpoint_total, max_bookmark_value, total_records

    def define_bookmark_filters(self, days_interval, last_datetime, now_datetime, attribution_window):
        """
        define the params from and to according to the filters provided in
         the bookmark_query_field_from and bookmark_query_field_to
        """
        if self.bookmark_query_field_from and self.bookmark_query_field_to:
            # days_interval from config date_window_size, default = 60; passed to function from sync
            if not days_interval:
                days_interval = 30

            last_dttm = strptime_to_utc(last_datetime)
            delta_days = (now_datetime - last_dttm).days
            if delta_days <= attribution_window:
                delta_days = attribution_window
                LOGGER.info("Start bookmark less than %s day attribution window.", attribution_window)
            elif delta_days >= 365:
                delta_days = 365
                LOGGER.warning("Start date or bookmark greater than 1 year maxiumum.")
                LOGGER.warning("Setting bookmark start to 1 year ago.")

            start_window = now_datetime - timedelta(days=delta_days)
            end_window = start_window + timedelta(days=days_interval)
            if end_window > now_datetime:
                end_window = now_datetime
        else:
            start_window = strptime_to_utc(last_datetime)
            end_window = now_datetime
            diff_sec = (end_window - start_window).seconds
            # round-up difference to days
            days_interval = math.ceil(diff_sec / (3600 * 24))

        return start_window, end_window, days_interval

    def sync(self, state, catalog, config, start_date):
        # the sync method common to all the streams which internally call methods depending on different endpoints

        bookmark_field = next(iter(self.replication_keys), None)
        project_timezone = config.get("project_timezone", "UTC")
        days_interval = int(config.get("date_window_size", "30"))
        attribution_window = int(config.get("attribution_window", "5"))

        #Update url if eu_residency is selected
        if str(config.get('eu_residency')).lower() == "true":
            if self.tap_stream_id == 'export':
                self.url = 'https://data-eu.mixpanel.com/api/2.0'
            else:
                self.url = 'https://eu.mixpanel.com/api/2.0'

        # Get the latest bookmark for the stream and set the last_integer/datetime
        last_datetime = self.get_bookmark(
            state, self.tap_stream_id, start_date)
        max_bookmark_value = last_datetime

        self.write_schema(catalog, self.tap_stream_id)

        # windowing: loop through date days_interval date windows from last_datetime to now_datetime
        tzone = pytz.timezone(project_timezone)
        now_datetime = datetime.now(tzone)
        end_date = config.get('end_date')

        if end_date:
            now_datetime = strptime_to_utc(end_date)

        start_window, end_window, days_interval = self.define_bookmark_filters(
            days_interval, last_datetime, now_datetime, attribution_window)
        # LOOP order: Date Windows, Parent IDs, Page
        # Initialize counter
        endpoint_total = 0  # Total for ALL: parents, date windows, and pages

        # Begin date windowing loop
        while start_window < now_datetime:
            # Initialize counters
            date_total = 0  # Total records for a date window
            parent_total = 0  # Total records for parent ID
            total_records = 0  # Total records for all pages
            record_count = 0  # Total processed for page

            params = self.params  # adds in endpoint specific, sort, filter params

            if self.bookmark_query_field_from and self.bookmark_query_field_to:
                # Request dates need to be normalized to project timezone or else errors may occur
                # Errors occur when from_date is > 365 days ago
                #   and when to_date > today (in project timezone)
                from_date = str(start_window.astimezone(tzone).date())
                to_date = str(end_window.astimezone(tzone).date())
                LOGGER.info("START Sync for Stream: %s", self.tap_stream_id)
                if self.bookmark_query_field_from:
                    LOGGER.info("Date window from: %s to %s", from_date, to_date)
                params[self.bookmark_query_field_from] = from_date
                params[self.bookmark_query_field_to] = to_date

            # funnels and cohorts have a parent endpoint with parent_data and parent_id_field
            if self.parent_path and self.parent_id_field:
                # API request data
                LOGGER.info("URL for Parent Stream %s: %s/%s",
                            self.tap_stream_id,
                            self.url,
                            self.parent_path)
                parent_data = self.client.request(
                    method="GET",
                    url=self.url,
                    path=self.parent_path,
                    endpoint="parent_data",
                )
            # Other endpoints (not funnels, cohorts): Simulate parent_data with single record
            else:
                parent_data = [{"id": "none"}]
                self.parent_id_field = "id"

            for parent_record in parent_data:
                parent_id = parent_record.get(self.parent_id_field)
                LOGGER.info('START: Stream: %s, parent_id: %s', self.tap_stream_id, parent_id)

                # pagination: loop thru all pages of data using next (if not None)
                page = 0  # First page is page=0, second page is page=1, ...
                offset = 0
                limit = 250  # Default page_size
                # Initialize counters
                parent_total = 0  # Total records for parent ID
                total_records = 0  # Total records for all pages
                record_count = 0  # Total processed for page

                session_id = 'initial'
                if self.pagination:
                    params['page_size'] = limit

                # Popped session_id and page number of last parents stream call.
                params.pop('session_id', None)
                params.pop('page', None)

                while offset <= total_records and session_id is not None:
                    if self.pagination and page != 0:
                        params['session_id'] = session_id
                        params['page'] = page

                    # querystring: Squash query params into string and replace [parent_id]
                    querystring = '&'.join(['%s=%s' % (key, value) for (key, value) in params.items()])
                    querystring = querystring.replace('[parent_id]', str(parent_id))

                    full_url = '{}/{}{}'.format(
                        self.url,
                        self.path,
                        '?{}'.format(querystring) if querystring else '')

                    LOGGER.info('URL for Stream %s: %s', self.tap_stream_id, full_url)

                    # API request data
                    # data = {}

                    parent_total, date_total, offset, page, session_id, endpoint_total, max_bookmark_value, total_records = self.get_and_transform_records(
                        querystring, project_timezone, max_bookmark_value, catalog, last_datetime, endpoint_total, limit, total_records, parent_total, record_count, page, offset, parent_record, date_total)
                   # End stream != 'export'
                LOGGER.info('FINISHED: Stream: %s, parent_id: %s', self.tap_stream_id, parent_id)
                LOGGER.info('Total records for parent: %s', parent_total)
                # End parent record loop
            LOGGER.info("FINISHED Sync for Stream: %s",self.tap_stream_id)
            if self.bookmark_query_field_from:
                LOGGER.info("Date window from: %s to %s",from_date, to_date)
            LOGGER.info("Total records for date window: %s", date_total)
            # Increment date window
            start_window = end_window
            next_end_window = end_window + timedelta(days=days_interval)
            if next_end_window > now_datetime:
                end_window = now_datetime
            else:
                end_window = next_end_window

            # Update the state with the max_bookmark_value for the stream
            if bookmark_field:
                self.write_bookmark(state, self.tap_stream_id, max_bookmark_value)
            # End date window loop
        # Return endpoint_total across all batches
        return endpoint_total


class Annotations(MixPanel):
    """
    List the annotations for a given date range.
    Docs: https://developer.mixpanel.com/reference/annotations
    """
    tap_stream_id = "annotations"
    key_properties = ["date"]
    path = "annotations"
    data_key = "annotations"
    bookmark_query_field_from = "from_date"
    bookmark_query_field_to = "to_date"
    replication_method = "FULL_TABLE"
    params = {}
    replication_keys = []


class CohortMembers(MixPanel):
    """
    The list endpoint returns all of the cohorts in a given project.
    The JSON formatted return contains the cohort name, id, count, description, creation date, and visibility for every cohort in the project.
    Docs: https://developer.mixpanel.com/reference/engage
    """
    tap_stream_id = "cohort_members"
    path = "engage"
    key_properties = ["cohort_id", "distinct_id"]
    params = {"filter_by_cohort": '{"id": [parent_id]}'}
    data_key = "results"
    pagination = True
    parent_path = "cohorts/list"
    parent_id_field = "id"
    replication_keys = []
    replication_method = "FULL_TABLE"
    bookmark_query_field_from = None
    bookmark_query_field_to = None


class Cohorts(MixPanel):
    """
    Takes a JSON object with a single key called id whose value is the cohort ID. behaviors and filter_by_cohort are mutually exclusive.
    Docs: https://developer.mixpanel.com/reference/cohorts
    """
    tap_stream_id = "cohorts"
    path = "cohorts/list"
    key_properties = ["id"]
    data_key = None
    replication_method = "FULL_TABLE"
    params = {}
    replication_keys = []
    bookmark_query_field_from = None
    bookmark_query_field_to = None


class Engage(MixPanel):
    """
    Query user profile data and return list of users that fit specified parameters.
    Docs: https://developer.mixpanel.com/reference/engage
    """
    tap_stream_id = "engage"
    path = "engage"
    data_key = "results"
    pagination = True
    key_properties = ["distinct_id"]
    replication_method = "FULL_TABLE"
    bookmark_query_field_from = None
    bookmark_query_field_to = None
    params = {}
    replication_keys = []


class Export(MixPanel):
    """
    Every data point sent to Mixpanel is stored as JSON in our data store.
    The raw export API allows you to download your event data as it is received and stored within Mixpanel,
    complete with all event properties (including distinct_id) and the exact timestamp the event was fired.
    Docs: https://developer.mixpanel.com/reference/export
    """
    tap_stream_id = "export"
    path = "export"
    data_key = "results"
    replication_keys = ["time"]
    url = "https://data.mixpanel.com/api/2.0"
    bookmark_query_field_from = "from_date"
    bookmark_query_field_to = "to_date"
    replication_method = "INCREMENTAL"
    params = {}

    def get_and_transform_records(
            self, querystring, project_timezone, max_bookmark_value, catalog, last_datetime, endpoint_total,
            limit, total_records, parent_total, record_count, page, offset, parent_record, date_total):
        """
        Get the records using the client get request and transform it using transform_records
        and return the max_bookmark_value
        """
        data = self.client.request_export(
            method='GET',
            url=self.url,
            path=self.path,
            params=querystring,
            endpoint=self.tap_stream_id)

        # time_extracted: datetime when the data was extracted from the API
        time_extracted = utils.now()
        transformed_data = []
        for record in data:
            if record and str(record) != '':
                # transform record and append to transformed_data array
                transformed_record = transform_record(record,
                                                      self.tap_stream_id,
                                                      project_timezone)
                transformed_data.append(transformed_record)

                # Check for missing keys
                for key in self.key_properties:
                    val = transformed_record.get(key)
                    if not val:
                        LOGGER.error('Error: Missing Key')
                        raise 'Missing Key'

                if len(transformed_data) == limit:
                    # Process full batch (limit = 250) records
                    #   and get the max_bookmark_value and record_count
                    max_bookmark_value, record_count = self.process_records(
                        catalog=catalog,
                        stream_name=self.tap_stream_id,
                        records=transformed_data,
                        time_extracted=time_extracted,
                        bookmark_field=next(iter(self.replication_keys), None),
                        max_bookmark_value=max_bookmark_value,
                        last_datetime=last_datetime)
                    LOGGER.info('Stream %s, batch processed %s records', self.tap_stream_id, record_count)

                    total_records = total_records + record_count
                    parent_total = parent_total + record_count
                    date_total = date_total + record_count
                    endpoint_total = endpoint_total + record_count
                    transformed_data = []
                    # End if (batch = limit 250)
                # End if record
            # End has export_data records loop

        # Process remaining, partial batch
        if len(transformed_data) > 0:
            max_bookmark_value, record_count = self.process_records(
                catalog=catalog,
                stream_name=self.tap_stream_id,
                records=transformed_data,
                time_extracted=time_extracted,
                bookmark_field=next(iter(self.replication_keys), None),
                max_bookmark_value=max_bookmark_value,
                last_datetime=last_datetime)
            LOGGER.info('Stream %s, batch processed %s records', self.tap_stream_id, record_count)

            total_records = total_records + record_count
            parent_total = parent_total + record_count
            date_total = date_total + record_count
            endpoint_total = endpoint_total + record_count
            # End if transformed_data

        # Export does not provide pagination; session_id = None breaks out of loop.
        session_id = None
        return parent_total, date_total, offset, page, session_id, endpoint_total, max_bookmark_value, total_records


class Funnels(MixPanel):
    """
    Get data for a funnel. funnel_id as a parameter to the API to get the funnel that you wish to get data for.
    Docs: https://developer.mixpanel.com/reference/funnels
    """
    tap_stream_id = "funnels"
    path = "funnels"
    key_properties = ["funnel_id", "date"]
    data_key = "data"
    date_dictionary = True
    bookmark_query_field_from = "from_date"
    bookmark_query_field_to = "to_date"
    replication_keys = ["datetime"]
    params = {"funnel_id": "[parent_id]", "unit": "day"}
    parent_path = "funnels/list"
    parent_id_field = "funnel_id"
    replication_method = "INCREMENTAL"


class Revenue(MixPanel):
    """
    Get the revenue data.
    """
    tap_stream_id = "revenue"
    path = "engage/revenue"
    key_properties = ["date"]
    data_key = "results"
    date_dictionary = True
    bookmark_query_field_from = "from_date"
    bookmark_query_field_to = "to_date"
    replication_keys = ["datetime"]
    params = {"unit": "day"}
    replication_method = "INCREMENTAL"


STREAMS = {
    "export": Export,
    "engage": Engage,
    "funnels": Funnels,
    "cohorts": Cohorts,
    "cohort_members": CohortMembers,
    "revenue": Revenue,
    "annotations": Annotations,
}



from math import ceil

import tap_tester.connections as connections
import tap_tester.runner as runner
import tap_tester.menagerie as menagerie
from tap_tester.logger import LOGGER

from base import TestMixPanelBase


class MixPanelPaginationAllFieldsTest(TestMixPanelBase):

    @staticmethod
    def name():
        return "mixpanel_pagination_all_fields_test"

    def pagination_test_run(self):
        """
        All Fields Test
        • and that when all fields are selected more than the automatic fields are replicated.
        • Verify no unexpected streams were replicated
        • Verify that more than just the automatic fields are replicated for each stream.
        • verify all fields for each stream are replicated
        • verify that the automatic fields are sent to the target


        Pagination Test
        • Verify that for each stream you can get multiple pages of data
        • Verify no duplicate pages are replicated
        • Verify no unexpected streams were replicated

        PREREQUISITE
        For EACH stream add enough data that you surpass the limit of a single
        fetch of data.  For instance if you have a limit of 250 records ensure
        that 251 (or more) records have been posted for that stream.
        """

        # Only following below 2 streams support pagination
        streams_to_test_all_fields = self.expected_streams()
        streams_to_test_pagination = {'engage', 'cohort_members'}

        expected_automatic_fields = self.expected_automatic_fields()
        conn_id = connections.ensure_connection(self)

        found_catalogs = self.run_and_verify_check_mode(conn_id)

        # table and field selection
        test_catalogs_all_fields = [catalog for catalog in found_catalogs
                                    if catalog.get('tap_stream_id') in streams_to_test_all_fields]

        self.perform_and_verify_table_and_field_selection(
            conn_id, test_catalogs_all_fields)

        # grab metadata after performing table-and-field selection to set expectations
        # used for asserting all fields are replicated
        stream_to_all_catalog_fields = dict()
        for catalog in test_catalogs_all_fields:
            stream_id, stream_name = catalog['stream_id'], catalog['stream_name']
            catalog_entry = menagerie.get_annotated_schema(conn_id, stream_id)
            fields_from_field_level_md = [md_entry['breadcrumb'][1]
                                          for md_entry in catalog_entry['metadata']
                                          if md_entry['breadcrumb'] != []]
            stream_to_all_catalog_fields[stream_name] = set(
                fields_from_field_level_md)

        record_count_by_stream = self.run_and_verify_sync(conn_id)

        actual_fields_by_stream = runner.examine_target_output_for_fields()

        synced_records = runner.get_records_from_target_output()

        # Verify no unexpected streams were replicated
        synced_stream_names = set(synced_records.keys())
        self.assertSetEqual(streams_to_test_all_fields, synced_stream_names)

        # All Fields Test
        for stream in streams_to_test_all_fields:
            with self.subTest(logging="Primary Functional Test", stream=stream):

                # expected values
                expected_all_keys = stream_to_all_catalog_fields[stream]
                expected_automatic_keys = expected_automatic_fields.get(
                    stream, set())

                # collect actual values
                messages = synced_records.get(stream)
                actual_all_keys = set()
                for message in messages['messages']:
                    if message['action'] == 'upsert':
                        actual_all_keys.update(set(message['data'].keys()))

                # verify that the automatic fields are sent to the target
                self.assertTrue(
                    actual_fields_by_stream.get(stream, set()).issuperset(
                        expected_automatic_keys),
                    msg="The fields sent to the target don't include all automatic fields")

                # Verify that more than just the automatic fields are replicated for each stream.
                if stream != "cohort_members":  # cohort_member has just 2 key and both are automatic
                    self.assertGreater(len(expected_all_keys),
                                       len(expected_automatic_keys))

                self.assertTrue(expected_automatic_keys.issubset(
                    expected_all_keys), msg=f'{expected_automatic_keys-expected_all_keys} is not in "expected_all_keys"')

                # As we can't find the below fields in the docs and also
                # it won't be generated by mixpanel APIs now so expected.
                if stream == "export":
                    expected_all_keys = expected_all_keys - {'labels', 'sampling_factor', 'dataset', 'mp_reserved_duration_s', 'mp_reserved_origin_end',
                                                             'mp_reserved_origin_start', 'mp_reserved_event_count'}

                # verify all fields for each stream are replicated
                if not stream == "engage": #Skip engage as it return records in random manner with dynamic fields.
                    self.assertSetEqual(expected_all_keys, actual_all_keys)

        # Pagination Test
        for stream in streams_to_test_pagination:
            with self.subTest(stream=stream):

                # expected values
                expected_primary_keys = self.expected_pks()[stream]

                # collect actual values
                messages = synced_records.get(stream)
                primary_keys_list = [tuple([message['data'][expected_pk] for expected_pk in expected_primary_keys])
                                     for message in messages['messages'] if message['action'] == 'upsert']

                # verify that we can paginate with all fields selected
                record_count_sync = record_count_by_stream.get(stream, 0)
                self.assertGreater(record_count_sync, self.API_LIMIT,
                                   msg="The number of records is not over the stream max limit")


                # Chunk the replicated records (just primary keys) into expected pages
                pages = []
                page_count = ceil(len(primary_keys_list) / self.API_LIMIT)
                page_size = self.API_LIMIT
                for page_index in range(page_count):
                    page_start = page_index * page_size
                    page_end = (page_index + 1) * page_size
                    pages.append(set(primary_keys_list[page_start:page_end]))

                # Verify by primary keys that data is unique for each page
                for current_index, current_page in enumerate(pages):
                    with self.subTest(current_page_primary_keys=current_page):

                        for other_index, other_page in enumerate(pages):
                            if current_index == other_index:
                                continue  # don't compare the page to itself

                            self.assertTrue(
                                current_page.isdisjoint(other_page), msg=f'other_page_primary_keys={other_page}'
                            )

    def test_run(self):
        # Pagination test for standard server
        self.eu_residency = False
        self.pagination_test_run()


        # Pagination test for EU residency server
        self.eu_residency = True
        self.pagination_test_run()



import singer

from tap_mixpanel.streams import STREAMS

LOGGER = singer.get_logger()


def update_currently_syncing(state, stream_name):
    """
    Currently syncing sets the stream currently being delivered in the state.
    If the integration is interrupted, this state property is used to identify
     the starting point to continue from.
    Reference: https://github.com/singer-io/singer-python/blob/master/singer/bookmarks.py#L41-L46
    """
    if (stream_name is None) and ('currently_syncing' in state):
        del state['currently_syncing']
    else:
        singer.set_currently_syncing(state, stream_name)
    singer.write_state(state)


def sync(client, config, catalog, state, start_date):
    """
    Get selected_streams from catalog, based on state last_stream
    last_stream = Previous currently synced stream, if the load was interrupted
    """
    last_stream = singer.get_currently_syncing(state)
    LOGGER.info('last/currently syncing stream: %s', last_stream)
    selected_streams = []
    for stream in catalog.get_selected_streams(state):
        selected_streams.append(stream.stream)
    LOGGER.info('selected_streams: %s', selected_streams)

    if not selected_streams:
        return

    # Loop through selected_streams
    for stream_name in selected_streams:
        LOGGER.info('START Syncing: %s', stream_name)
        update_currently_syncing(state, stream_name)
        stream_obj = STREAMS[stream_name](client)
        endpoint_total = stream_obj.sync(
            catalog=catalog,
            state=state,
            config=config,
            start_date=start_date
        )

        update_currently_syncing(state, None)
        LOGGER.info('FINISHED Syncing: %s, Total endpoint records: %s', stream_name, endpoint_total)



import unittest
from unittest import mock
from tap_mixpanel.streams import  Revenue, Export
from tap_mixpanel.client import MixpanelClient
from tap_mixpanel.__init__ import main

EU_CONFIG = {
  "api_secret": "dummy_secret",
  "date_window_size": "30",
  "attribution_window": "5",
  "project_timezone": "Europe/Amsterdam",
  "select_properties_by_default": "true",
  "start_date": "2020-02-01T00:00:00Z",
  "user_agent": "tap-mixpanel <api_user_email@your_company.com>",
  "eu_residency": True,
  "end_date": "2020-03-02T00:00:00Z"
}
STANDARD_CONFIG = {
    "api_secret": "dummy_secret",
    "date_window_size": "30",
    "attribution_window": "5",
    "project_timezone": "Europe/Amsterdam",
    "select_properties_by_default": "true",
    "start_date": "2020-02-01T00:00:00Z",
    "user_agent": "tap-mixpanel <api_user_email@your_company.com>",
    "eu_residency": False,
    "end_date": "2020-03-02T00:00:00Z"
}

class MockStream():
    def __init__(self, stream):
        self.stream = stream

class MockCatalog():
    def __init__(self, name):
        self.name = name

    def get_selected_streams(self, state):
        return [MockStream(self.name)]

class MockParseArgs():
    def __init__(self, state, discover, config):
        self.state = state
        self.discover = discover
        self.config = config

class Mockresponse:
    def __init__(self, resp, status_code, content=[""], headers=None, raise_error=False, text={}):
        self.json_data = resp
        self.status_code = status_code
        self.content = content
        self.headers = headers
        self.raise_error = raise_error
        self.text = text
        self.reason = "error"

    def prepare(self):
        return (self.json_data, self.status_code, self.content, self.headers, self.raise_error)

    def json(self):
        return self.text

class TestMixpanelSupportEuEndpoints(unittest.TestCase):

    @mock.patch("tap_mixpanel.client.MixpanelClient.request")
    @mock.patch("tap_mixpanel.streams.MixPanel.write_bookmark")
    @mock.patch("tap_mixpanel.streams.MixPanel.write_schema")
    def test_support_eu_endpoints_except_export(self, mock_write_schema, mock_write_bookmark, mock_request):
        mock_request.return_value = {}
        mock_write_schema.return_value = ''
        mock_write_bookmark.return_value = ''

        state = {}
        catalog = MockCatalog('revenue')

        client = MixpanelClient('','','')
        revenue_obj = Revenue(client)
        revenue_obj.sync(catalog=catalog, state=state, config=EU_CONFIG, start_date="2020-02-01T00:00:00Z")

        mock_request.assert_called_with(method='GET', url='https://eu.mixpanel.com/api/2.0', path='engage/revenue',
                                        params='unit=day&from_date=2020-02-01&to_date=2020-03-02', endpoint='revenue')

        revenue_obj = Revenue(client)
        revenue_obj.sync(catalog=catalog,state=state, config=STANDARD_CONFIG, start_date="2020-02-01T00:00:00Z")

        mock_request.assert_called_with(method='GET', url='https://mixpanel.com/api/2.0', path='engage/revenue',
                                        params='unit=day&from_date=2020-02-01&to_date=2020-03-02', endpoint='revenue')


    @mock.patch("tap_mixpanel.client.MixpanelClient.request_export")
    @mock.patch("tap_mixpanel.streams.MixPanel.write_bookmark")
    @mock.patch("tap_mixpanel.streams.MixPanel.write_schema")
    def test_support_export_eu_endpoint(self, mock_write_schema, mock_write_bookmark, mock_request_export):
        mock_request_export.return_value = {}
        mock_write_schema.return_value = ''
        mock_write_bookmark.return_value = ''

        state = {}
        catalog = MockCatalog('export')

        client = MixpanelClient('','','')
        export_obj = Export(client)
        export_obj.sync(catalog=catalog,state=state, config=EU_CONFIG, start_date="2020-02-01T00:00:00Z")

        mock_request_export.assert_called_with(method='GET', url='https://data-eu.mixpanel.com/api/2.0', path='export',
                                        params='from_date=2020-02-01&to_date=2020-03-02', endpoint='export')

        export_obj = Export(client)
        export_obj.sync(catalog=catalog,state=state, config=STANDARD_CONFIG, start_date="2020-02-01T00:00:00Z")

        mock_request_export.assert_called_with(method='GET', url='https://data.mixpanel.com/api/2.0', path='export',
                                        params='from_date=2020-02-01&to_date=2020-03-02', endpoint='export')


    @mock.patch("requests.Session.request")
    @mock.patch("singer.utils.parse_args")
    @mock.patch("tap_mixpanel.__init__.do_discover")
    def test_support_eu_endpoint_in_discover(self, mock_discover, mock_parse_args, mock_request):

        mock_request.return_value = Mockresponse("", status_code=200)
        mock_discover.return_value = ''
        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=EU_CONFIG)
        r = main()

        header = {
            'User-Agent': 'tap-mixpanel <api_user_email@your_company.com>',
            'Accept': 'application/json',
            'Authorization': 'Basic ZHVtbXlfc2VjcmV0'
        }
        mock_request.assert_called_with('GET','https://eu.mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=header, timeout=300)

        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=STANDARD_CONFIG)
        r = main()

        mock_request.assert_called_with('GET','https://mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=header, timeout=300)



import pytest
from tap_mixpanel.client import MixpanelClient


@pytest.fixture
def mixpanel_client():
    # Support of request_timeout have been added. So, now MixpanelClient accept request_timeout parameter which is mandatory
    mixpanel_client = MixpanelClient('API_SECRET', api_domain="mixpanel.com", request_timeout=1) # Pass extra request_timeout parameter
    mixpanel_client._MixpanelClient__verified = True
    return mixpanel_client



import os
import json
from singer import metadata
from tap_mixpanel.streams import STREAMS
import singer

LOGGER = singer.get_logger()

# Reference:
# https://github.com/singer-io/getting-started/blob/master/docs/DISCOVERY_MODE.md#Metadata

def get_abs_path(path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)


def get_schema(client, properties_flag, stream_name):
    schema_path = get_abs_path('schemas/{}.json'.format(stream_name))

    with open(schema_path) as file:
        schema = json.load(file)

    # Set whether to allow additional properties for engage and export endpoints
    # Event and Engage properties are dynamic and depend on the properties provided on upload,
    #   when the Event or Engage (user/person) was created.
    # Depending on the tap config parameter select_properties_by_default,
    #   the json schema should allow additional properties (additionalProperties = true).
    if stream_name in ('engage', 'export') and str(properties_flag).lower() == 'true':
        schema['additionalProperties'] = True
    else:
        schema['additionalProperties'] = False

    if stream_name == 'engage':
        properties = client.request(
            method='GET',
            url='https://{}/api/2.0'.format(client.__api_domain),
            path='engage/properties',
            params={'limit': 2000},
            endpoint='engage_properties')
        if properties.get('status') == 'ok':
            results = properties.get('results', {})
            for key, val in results.items():
                if key[0:1] == '$':
                    new_key = 'mp_reserved_{}'.format(key[1:])
                else:
                    new_key = key

                # property_type: string, number, boolean, datetime, object, list
                # Reference:
                # https://help.mixpanel.com/hc/en-us/articles/115004547063-Properties-Supported-Data-Types
                property_type = val.get('type')

                types = {
                    'boolean': {
                        'type': ['null', 'boolean']
                    },
                    'number': {
                        'type': ['null', 'string'],
                        'format': 'singer.decimal'
                    },
                    'datetime': {
                        'type': ['null', 'string'],
                        'format': 'date-time'
                    },
                    'object': {
                        'type': ['null', 'object'],
                        'additionalProperties': True
                    },
                    'list': {
                        'type': ['null', 'array'],
                        'required': False,
                        'items': {}
                    },
                    'string': {
                        'type': ['null', 'string']
                    }
                }

                if property_type in types.keys():
                    # Make the types a list containing all types starting with the one returned to us by the API
                    this_type = [types.pop(property_type)]
                    this_type += list(types.values())

                else:
                    this_type = list(types.values())


                schema['properties'][new_key] = {'anyOf': this_type}

    if stream_name == 'export':
        # Event properties endpoint:
        #  https://developer.mixpanel.com/docs/data-export-api#section-hr-span-style-font-family-courier-top-span
        results = client.request(
            method='GET',
            url='https://{}/api/2.0'.format(client.__api_domain),
            path='events/properties/top',
            params={'limit': 2000},
            endpoint='event_properties')
        for key, val in results.items():
            if key[0:1] == '$':
                new_key = 'mp_reserved_{}'.format(key[1:])
            else:
                new_key = key

            # string ONLY for event properties (no other datatypes)
            # Reference: https://help.mixpanel.com/hc/en-us/articles/360001355266-Event-Properties#field-size-character-limits-for-event-properties
            schema['properties'][new_key] = {
                'type': ['null', 'string']
            }

    return schema

def get_schemas(client, properties_flag):
    schemas = {}
    field_metadata = {}

    for stream_name, stream_metadata in STREAMS.items():
        # When the client detects disable_engage_endpoint, skip discovering the stream
        if stream_name == 'engage' and client.disable_engage_endpoint:
            LOGGER.warning('Mixpanel returned a 402 indicating the Engage endpoint and stream is unavailable. Skipping.')
            continue

        schema = get_schema(client, properties_flag, stream_name)

        schemas[stream_name] = schema
        mdata = metadata.new()

        # Documentation:
        # https://github.com/singer-io/getting-started/blob/master/docs/DISCOVERY_MODE.md#singer-python-helper-functions
        # Reference:
        # https://github.com/singer-io/singer-python/blob/master/singer/metadata.py#L25-L44
        mdata = metadata.get_standard_metadata(
            schema=schema,
            key_properties=stream_metadata.key_properties,
            valid_replication_keys=stream_metadata.replication_keys,
            replication_method=stream_metadata.replication_method
        )

        mdata = metadata.to_map(mdata)

        if stream_metadata.replication_keys:
                mdata = metadata.write(
                    mdata,
                    ('properties', stream_metadata.replication_keys[0]),
                    'inclusion',
                    'automatic')

        mdata = metadata.to_list(mdata)

        field_metadata[stream_name] = mdata

    return schemas, field_metadata



import unittest
from unittest import mock
from tap_mixpanel.__init__ import main

CONFIG = {
    "api_secret": "dummy_secret",
    "date_window_size": "30",
    "attribution_window": "5",
    "project_timezone": "Europe/Amsterdam",
    "select_properties_by_default": "true",
    "start_date": "2020-02-01T00:00:00Z",
    "user_agent": "tap-mixpanel <api_user_email@your_company.com>",
    "eu_residency": False,
    "end_date": "2020-03-02T00:00:00Z"
}

REQUEST_TIMEOUT = 300
REQUEST_TIMEOUT_FLOAT = 300.0

class MockParseArgs():
    """
    Mocked MockParseArgs class with custom state, discover, config attributes to pass unit test cases.
    """
    def __init__(self, state, discover, config):
        self.state = state
        self.discover = discover
        self.config = config

class Mockresponse:
    """
    Mocked standard HTTPResponse.
    """
    def __init__(self, resp, status_code):
        self.json_data = resp
        self.status_code = status_code

HEADER = {
        'User-Agent': 'tap-mixpanel <api_user_email@your_company.com>',
        'Accept': 'application/json',
        'Authorization': 'Basic ZHVtbXlfc2VjcmV0'
    }
   
@mock.patch("requests.Session.request", return_value = Mockresponse("", status_code=200))
@mock.patch("singer.utils.parse_args")
@mock.patch("tap_mixpanel.__init__.do_discover", return_value = '')
class TestMixpanelRequestTimeoutParameterValue(unittest.TestCase):
    """Test that tap handles different type of request_timeout parameter values"""
    def test_request_timeout_for_none_param_value(self, mock_discover, mock_parse_args, mock_request):
        """Test that tap handles none value of request_timeout parameter"""
        config = CONFIG.copy()
        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=config)
        r = main()

        # Verify that request method called with expected parameter value when"request_timeout" is None
        mock_request.assert_called_with('GET','https://mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=HEADER, timeout=REQUEST_TIMEOUT)
    
    def test_request_timeout_for_empty_param_value(self, mock_discover, mock_parse_args, mock_request):
        """Test that tap handles empty value of request_timeout parameter"""
        config = CONFIG.copy()
        config['request_timeout'] = ""
        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=config)
        r = main()

        # Verify that request method called with expected parameter value when"request_timeout" is empty string
        mock_request.assert_called_with('GET','https://mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=HEADER, timeout=REQUEST_TIMEOUT)
    
    def test_request_timeout_for_string_param_value(self, mock_discover, mock_parse_args, mock_request):
        """Test that tap handles string value of request_timeout parameter"""
        config = CONFIG.copy()
        config['request_timeout'] = "100"
        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=config)
        r = main()

        # Verify that request method called with expected parameter value when"request_timeout" is string
        mock_request.assert_called_with('GET','https://mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=HEADER, timeout=100.0)
    
    def test_request_timeout_for_int_param_value(self, mock_discover, mock_parse_args, mock_request):
        """Test that tap handles int value of request_timeout parameter"""
        config = CONFIG.copy()
        config['request_timeout'] = 200
        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=config)
        r = main()

        # Verify that request method called with expected parameter value when"request_timeout" is int
        mock_request.assert_called_with('GET','https://mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=HEADER, timeout=200.0)
        
    def test_request_timeout_for_float_param_value(self, mock_discover, mock_parse_args, mock_request):
        """Test that tap handles float value of request_timeout parameter"""
        config = CONFIG.copy()
        config['request_timeout'] = REQUEST_TIMEOUT_FLOAT
        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=config)
        r = main()

        # Verify that request method called with expected parameter value when"request_timeout" is float
        mock_request.assert_called_with('GET','https://mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=HEADER, timeout=REQUEST_TIMEOUT_FLOAT)
        
    def test_request_timeout_for_zero_int_param_value(self, mock_discover, mock_parse_args, mock_request):
        """Test that tap handles int 0 value of request_timeout parameter"""
        config = CONFIG.copy()
        config['request_timeout'] = 0
        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=config)
        r = main()

        # Verify that request method called with expected parameter value when"request_timeout" is int 0
        mock_request.assert_called_with('GET','https://mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=HEADER, timeout=REQUEST_TIMEOUT)
  
    def test_request_timeout_for_zero_string_param_value(self, mock_discover, mock_parse_args, mock_request):
        """Test that tap handles string 0 value of request_timeout parameter"""
        config = CONFIG.copy()
        config['request_timeout'] = "0"
        mock_parse_args.return_value = MockParseArgs(state = {}, discover = True, config=config)
        r = main()

        # Verify that request method called with expected parameter value when"request_timeout" is string 0
        mock_request.assert_called_with('GET','https://mixpanel.com/api/2.0/engage', allow_redirects=True,
                                        headers=HEADER, timeout=REQUEST_TIMEOUT)

